import py
import textwrap
from contextlib import contextmanager
from pypytools import compat

class Code(object):

    def __init__(self, pyx=False):
        self.pyx = pyx
        self._lines = []
        self._indentation = 0
        self._globals = compat.newdict('module')
        self.global_scope = Scope(self)
        #
        self.new_scope = self.global_scope.new_scope
        self.w = self.global_scope.w
        self.ww = self.global_scope.ww
        self.block = self.global_scope.block
        self.def_ = self.global_scope.def_
        self.cpdef_ = self.global_scope.cpdef_
        self.cdef_ = self.global_scope.cdef_
        self.cdef_var = self.global_scope.cdef_var

    def build(self):
        return '\n'.join(self._lines)

    def compile(self):
        src = self.build()
        src = py.code.Source(src)
        co = src.compile()
        exec(co, self._globals)

    def __getitem__(self, name):
        return self._globals[name]

    def __setitem__(self, name, value):
        self._globals[name] = value

    def new_global(self, name, value):
        if name in self._globals:
            ext_value = self._globals[name]
            if value == ext_value:
                return name # nothing to do
            else:
                # try to find an unique name
                name = self._new_global_name(name)
        self._globals[name] = value
        return name

    def _new_global_name(self, name):
        i = 1
        while True:
            tryname = '%s__%d' % (name, i)
            if tryname not in self._globals:
                return tryname
            i += 1

    def args(self, varnames, args=None, kwargs=None):
        """
        Format a string representing an argument list to be used in calls.
        """
        return self._format_args(varnames, args, kwargs, mode='args')

    def params(self, varnames, args=None, kwargs=None):
        """
        Format a string representing a parameter list to be used in function definitions
        """
        return self._format_args(varnames, args, kwargs, mode='params')

    def _format_args(self, varnames, args, kwargs, mode):
        def typevar(varname):
            # Cython gets confused if we use e.g. 'void' as a varname; we need
            # to explicitly use "object void" to make it working. For
            # simplicity, we simply add an explicit "object" type to every
            # argument, when in pyx mode
            if mode == 'params' and self.pyx:
                return 'object %s' % varname
            return varname
        #
        def getvar(varname):
            if isinstance(varname, tuple):
                name, default = varname
                return '%s=%s' % (typevar(name), default)
            return typevar(varname)
        #
        parts = [getvar(v) for v in varnames]
        if args is not None:
            parts.append(args)
        if kwargs is not None:
            parts.append(kwargs)
        return ', '.join(parts)

    def call(self, funcname, varnames, args=None, kwargs=None):
        arglist = self.args(varnames, args, kwargs)
        return '{funcname}({arglist})'.format(funcname=funcname,
                                              arglist=arglist)


class Scope(object):

    # use __slots__ to ensure that we can have a __code attribute without
    # having it in __dict__: we want __dict__ to be empty by default, so that
    # we know when to apply formatting
    __slots__ = ['__code', '__dict__']

    def __init__(self, code, **kwargs):
        self.__code = code
        self.__dict__.update(kwargs)

    def _kwargs(self, kwargs):
        if self.__dict__:
            d = self.__dict__.copy()
            d.update(kwargs)
            return d
        return kwargs

    def new_scope(self, **kwargs):
        kwargs = self._kwargs(kwargs)
        return Scope(self.__code, **kwargs)

    def format(self, s, **kwargs):
        kwargs = self._kwargs(kwargs)
        if kwargs:
            s = s.format(**kwargs)
        return s

    def w(self, *parts, **kwargs):
        s = ' '.join(parts)
        s = self.format(s, **kwargs)
        self.__code._lines.append(' ' * self.__code._indentation + s)

    def ww(self, s, **kwargs):
        """
        Like code.w, but dedent the code first. Useful with triple-quoted strings
        """
        s = textwrap.dedent(s[1:])
        for line in s.splitlines():
            self.w(line, **kwargs)

    @contextmanager
    def block(self, s=None, autopass=True, **kwargs):
        ns = self.new_scope(**kwargs)
        if s is not None:
            ns.w(s)
        self.__code._indentation += 4
        n = len(self.__code._lines)
        yield ns
        if autopass and n == len(self.__code._lines):
            # we didn't write anything in the block, so automatically put a 'pass'
            ns.w('pass')
        self.__code._indentation -= 4

    def def_(self, funcname, varnames, args=None, kwargs=None):
        paramlist = self.__code.params(varnames, args, kwargs)
        return self.block('def {funcname}({paramlist}):',
                          funcname=funcname, paramlist=paramlist)

    def cdef_(self, funcname, varnames, args=None, kwargs=None):
        """
        In pyx mode, this emits a cdef block; in py mode, it's the very same as
        def_
        """
        paramlist = self.__code.params(varnames, args, kwargs)
        cpdef = self.__code.pyx and 'cdef' or 'def'
        return self.block('{cpdef} {funcname}({paramlist}):',
                          cpdef=cpdef, funcname=funcname, paramlist=paramlist)

    def cpdef_(self, funcname, varnames, args=None, kwargs=None):
        """
        In pyx mode, this emits a cpdef block; in py mode, it's the very same as
        def_
        """
        paramlist = self.__code.params(varnames, args, kwargs)
        cpdef = self.__code.pyx and 'cpdef' or 'def'
        return self.block('{cpdef} {funcname}({paramlist}):',
                          cpdef=cpdef, funcname=funcname, paramlist=paramlist)

    def cdef_var(self, typename, varname, default=None):
        if not self.__code.pyx:
            return
        if default is None:
            self.w('cdef {typename} {varname}', typename=typename, varname=varname)
        else:
            self.w('cdef {typename} {varname} = {default}',
                   typename=typename, varname=varname, default=default)



import sys
import inspect
import dis
from pypytools.unroll import unroll, Closure
from pypytools.util import PY3

if PY3:
    from io import StringIO
else:
    from cStringIO import StringIO


def test_make_closure():
    def foo():
        return x

    cl = Closure(foo, x=42)
    foo2 = cl.make()
    assert foo2() == 42

def test_inject_value():
    @unroll(x=42)
    def foo():
        return x
    assert foo() == 42

def test_existing_closure():
    y = 10
    @unroll(x=52)
    def foo():
        return x-y
    assert foo() == 42


def test_existing_closure_override():
    x = 10
    @unroll(x=42)
    def foo():
        return x
    assert foo() == 42


def test_getsource():
    def foo():
        pass
    #
    foo2 = unroll()(foo)
    src = inspect.getsource(foo)
    src2 = inspect.getsource(foo2)
    assert src == src2


def test_unrolling(monkeypatch):
    @unroll(items=(1, 2, 3))
    def foo():
        x = 0
        for i in items:
            x += i
        return x
    assert foo() == 6
    #
    stdout = StringIO()
    monkeypatch.setattr(sys, 'stdout', stdout)
    dis.dis(foo)
    monkeypatch.undo()
    assert 'FOR_ITER' not in stdout.getvalue()

def test_iterable(monkeypatch):
    items = iter([1, 2, 3])
    @unroll(items=items)
    def foo():
        x = 0
        for i in items:
            x += i
        return x
    assert foo() == 6
    #
    stdout = StringIO()
    monkeypatch.setattr(sys, 'stdout', stdout)
    dis.dis(foo)
    monkeypatch.undo()
    assert 'FOR_ITER' not in stdout.getvalue()


def my_global():
    return 42

def test_globals():
    @unroll()
    def foo():
        return my_global()
    assert foo() == 42




import pypytools
assert pypytools.IS_PYPY
from _numpypy.multiarray import *
from _numpypy.umath import *

from math import e, pi
PZERO = float('0.0')
NZERO = float('-0.0')
PINF = float('inf')
NINF = float('-inf')
NAN = float('nan')
euler_gamma = 0.577215664901532860606512090082402431 # from npy_math.h
Inf = inf = infty = Infinity = PINF
nan = NaN = NAN

def asarray(a, dtype=None, order=None):
    return array(a, dtype, copy=False, order=order)

def asanyarray(a, dtype=None, order=None):
    if isinstance(a, ndarray):
        return a
    return asarray(a, dtype, order)

def array_equal(a1, a2):
    try:
        a1, a2 = asarray(a1), asarray(a2)
    except:
        return False
    if a1.shape != a2.shape:
        return False
    return bool(asarray(a1 == a2).all())

def not_equal(a, b):
    return asarray(a) != b

def subtract(a, b):
    return asarray(a) - b

# generate wrappers for ndarray's methods
def make_wrappers():
    from pypytools.codegen import Code
    code = Code()
    names = ['all', 'any', 'argmax', 'argmin', 'argsort', 'choose', 'clip',
             'copy', 'cumprod', 'cumsum', 'diagonal', 'max', 'min',
             'nonzero', 'prod', 'ptp', 'put', 'ravel', 'repeat', 'reshape',
             'round', 'squeeze', 'sum', 'swapaxes', 'transpose']
    #
    for name in names:
        ns = code.new_scope(name=name)
        with code.def_(name, ['myarray'], '*args', '**kwargs'):
            ns.w('return myarray.{name}(*args, **kwargs)')
    #
    code.compile()
    gbl = globals()
    for name in names:
        gbl[name] = code[name]

make_wrappers()
del make_wrappers



import pypytools
assert pypytools.IS_PYPY
from _numpypy.multiarray import dtype

_types = ['bool_', 'int8', 'uint8', 'int16', 'uint16', 'int32',
          'uint32', 'int64', 'uint64', 'float16', 'float32',
          'float64', 'complex64', 'complex128', 'object']

for _typename in _types:
    globals()[_typename] = dtype(_typename).type



import time
from pytest import approx
from freezegun import freeze_time
from pypytools.gc.uniform import UniformGcStrategy

class FakeGcCollectStats(object):

    def __init__(self, major_is_done):
        self.major_is_done = major_is_done

class TestUniformGcStrategy(object):

    def new(self, initial_mem=0, **kwds):
        # we call __new__ and __init__ separately so that we can patch **kwds
        # before they are used in the __init__ (e.g., MIN_TARGET)
        s = UniformGcStrategy.__new__(UniformGcStrategy)
        s.__dict__.update(**kwds)
        s.__init__(initial_mem)
        return s

    def fakestats(self, is_done=False):
        return FakeGcCollectStats(is_done)

    def test_target_allocated_mem(self):
        s = self.new(MAJOR_COLLECT=1.8, MIN_TARGET=50)
        assert s.target_mem == 50
        assert s.target_allocated_mem == 50

        s.start_another_major(mem=100)
        assert s.target_mem == 180 # 100*1.8
        assert s.target_allocated_mem == 80

        s.start_another_major(mem=20)
        assert s.target_mem == 50 # MIN_TARGET
        assert s.target_allocated_mem == 30

    def test_estimated_gc(self):
        s = self.new(MAJOR_COLLECT=1.8, MIN_TARGET=1)
        # the last GC took 1s for a target_mem of 100
        s.target_mem = 100
        s.gc_cumul_t = 1.0

        s.start_another_major(mem=100)
        assert s.target_mem == 180
        assert s.gc_estimated_t == 1.8

    def test_alloc_rate(self):
        with freeze_time('2018-01-01') as freezer:
            s = self.new(initial_mem=100)
            freezer.tick(0.5)           # 0.5 second
            s.tick(mem=150)             # delta_mem == 50
            assert s.alloc_rate == 100  # 50/0.5 bytes/s

            freezer.tick(2)             # 2 seconds
            s.tick(mem=250)             # delta_mem == 100
            assert s.alloc_rate == 75   # because of the average

            freezer.tick(1)
            s.tick(mem=100)             # negative delta_mem
            assert s.alloc_rate == 38   # capped at 1

    def test_alloc_rate_nonzero(self):
        s = self.new(initial_mem=100)
        s.tick(mem=90)
        assert s.alloc_rate == 1

    def test_get_time_for_next_step(self):
        s = self.new(initial_mem=0)
        # time to allocate 900 bytes:  9 s
        # time estimated for the GC:   1 s
        # total estimated time:       10 s
        # GC / total:                 10%
        s.target_allocated_mem = 900.0
        s.allocated_mem = 0
        s.alloc_rate = 100.0 # bytes/s
        s.gc_estimated_t = 1
        # If the last step took 0.01 s, I wait for 0.09 to keep the expected
        # GC/total ration
        s.gc_last_step_t = 42
        s.gc_last_step_duration = 0.01
        t = s.get_time_for_next_step()
        assert t == 42.09

        # the time doesn't change if I allocate at the expected rate
        s.allocated_mem = 300
        s.gc_last_step_t = 42.03
        t = s.get_time_for_next_step()
        assert t == 42.09

        # the result changes accordingly to the allocation rate: if I allocate
        # slower, I have more time to finish the collection
        s.allocated_mem = 0
        s.alloc_rate = 10.0
        s.gc_last_step_t = 42
        t = s.get_time_for_next_step()
        assert t == 42.9

        # if I allocate faster, I need to hurry up
        s.alloc_rate = 1000.0
        t = s.get_time_for_next_step()
        assert t == 42.009

    def test_emergency_delay(self):
        # we are using too much memory and we have not finished the GC yet
        s = self.new(EMERGENCY_DELAY=3)
        s.gc_last_step_t = 39
        s.allocated_mem = 100
        s.target_allocated_mem = 100
        assert s.get_time_for_next_step() == 42

    def test_allocated_mem(self):
        s = self.new(initial_mem=500)
        assert s.allocated_mem == 0
        
        s.tick(mem=600)
        assert s.allocated_mem == 100
        s.tick(mem=700)
        assert s.allocated_mem == 200

        # simulate a sweep step, in which the mem usage decrease
        s.record_gc_step(mem=650, duration=0, stats=self.fakestats())
        s.tick(mem=690) # mem is back at 690, +40 bytes
        assert s.allocated_mem == 240

    def test_should_collect(self):
        with freeze_time('2018-01-01') as freezer:
            s = self.new(initial_mem=0)
            # with the following params and an alloc_rate of 100 bytes/s, the
            # GC takes an estimated 10% of the time
            s.target_allocated_mem = 900.0
            s.gc_estimated_t = 1
            s.gc_last_step_duration = 0.01

            # so, we expect to run 9 iterations before doing one step
            mem = 0
            i = 0
            while True:
                freezer.tick(0.01)
                i += 1
                mem += 1
                should_collect = s.tick(mem=mem)
                assert s.alloc_rate == approx(100) # floating point rounding :(
                assert s.last_t == time.time()
                assert s.last_mem == mem
                if should_collect:
                    break
            assert i == 9

    def test_record_gc_step(self):
        with freeze_time('2018-01-01') as freezer:
            s = self.new(initial_mem=0,
                         MAJOR_COLLECT=1.8,
                         MIN_TARGET=50)
            assert s.n_majors == 0

            t = time.time()
            s.record_gc_step(100, 2, self.fakestats(is_done=False))
            s.record_gc_step(110, 3, self.fakestats(is_done=False))
            s.record_gc_step(120, 1, self.fakestats(is_done=False))
            assert s.gc_cumul_t == 2+3+1
            assert s.gc_steps == 3
            assert s.last_mem == 120
            assert s.last_t == t
            assert s.gc_last_step_t == t

            freezer.tick(1)
            s.record_gc_step(80, 1, self.fakestats(is_done=True))
            assert s.n_majors == 1
            assert s.gc_cumul_t == 0
            assert s.gc_steps == 0
            assert s.last_mem == 80
            assert s.last_t == t+1
            assert s.gc_last_step_t == t+1
            assert s.target_allocated_mem == 80*0.8

    def test_adjust_gc_estimated_t(self):
        s = self.new(initial_mem=0, ESTIMATED_OVERFLOW_FACTOR=1.2)
        s.gc_estimated_t = 10.0
        stats = self.fakestats(is_done=False)
        s.record_gc_step(mem=100, duration=9, stats=stats)
        assert s.gc_estimated_t == 10.0

        # gc_cumul_t => 11.5, adjust the estimate
        s.record_gc_step(mem=100, duration=1.5, stats=stats)
        assert s.gc_estimated_t == 12.0

        # if gc_cumul_t is much bigger than the estimate, a single
        # *=ESTIMATED_OVERFLOW_FACTOR is not enouch; keep multiplying until
        # the estimate is bigger than the measured
        # gc_cumul_t ==> 14.5
        s.record_gc_step(mem=100, duration=4.5, stats=stats)
        assert s.gc_estimated_t == 12.0 * 1.2 * 1.2



"""
Improve PyPy's GC hooks and make it possible to have multiple callbacks
"""
from pypytools import IS_PYPY
import gc
import types

class GcHooks(object):
    """
    Main entry point to use multiple GC hooks. Make a subclass, override the
    desided handlers and call enable() and disable() appropriately.
    """

    on_gc_minor = None
    on_gc_collect_step = None
    on_gc_collect = None

    def enable(self):
        return MultiHook.get().add(self)

    def disable(self):
        return MultiHook.get().remove(self)


class MultiHook(object):
    """
    This is the actual class which is installed as GC hooks. It is not meant
    to be used directly, but to be manipulates by GcHooks()
    """

    # MultiHook is supposed to be a singleton (apart from tests). Use
    # MultiHook.get() to access it
    _instance = None

    @classmethod
    def get(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __new__(cls):
        # some magic to make MultiHook not crashing on CPython. If we don't
        # have gc.hook, MultiHook() returns an instance of FakeMultiHook
        self = object.__new__(cls)
        if not hasattr(gc, 'hooks'):
            self.__class__ = FakeMultiHook
        return self

    def __init__(self):
        self.hooks = [] # list of GcHooks instances
        self._update_callbacks()

    def add(self, gchooks):
        self.hooks.append(gchooks)
        self._update_callbacks()
        return True

    def remove(self, gchooks):
        self.hooks.remove(gchooks)
        self._update_callbacks()
        return True

    def _check_other_hooks(self):
        # Safety check to avoid disabling other hooks by mistake: the only
        # permitted hooks are the ones installed by us, i.e. bound methods of
        # "self". If we find extraneous hooks, complain
        def check(m):
            return (m is None or
                    (isinstance(m, types.MethodType) and m.__self__ is self))

        if not (check(gc.hooks.on_gc_minor) and
                check(gc.hooks.on_gc_collect_step) and
                check(gc.hooks.on_gc_collect)):
            raise ValueError("It seems other GC hooks are already installed. "
                             "Consider to use multihook everywhere.")

    def _update_callbacks(self):
        self._check_other_hooks()
        self.minor_callbacks = []
        self.collect_step_callbacks = []
        self.collect_callbacks = []
        for h in self.hooks:
            cb = getattr(h, 'on_gc_minor', None)
            if cb:
                self.minor_callbacks.append(cb)

            cb = getattr(h, 'on_gc_collect_step', None)
            if cb:
                self.collect_step_callbacks.append(cb)

            cb = getattr(h, 'on_gc_collect', None)
            if cb:
                self.collect_callbacks.append(cb)

        if self.minor_callbacks:
            gc.hooks.on_gc_minor = self.on_gc_minor

        if self.collect_step_callbacks:
            gc.hooks.on_gc_collect_step = self.on_gc_collect_step

        if self.collect_callbacks:
            gc.hooks.on_gc_collect = self.on_gc_collect
    
    def on_gc_minor(self, stats):
        for cb in self.minor_callbacks:
            cb(stats)

    def on_gc_collect_step(self, stats):
        for cb in self.collect_step_callbacks:
            cb(stats)

    def on_gc_collect(self, stats):
        for cb in self.collect_callbacks:
            cb(stats)


class FakeMultiHook(object):
    """
    This is the fake class which is used on CPython. See MultiHook.__new__
    """

    def is_working(self):
        return False
    
    def add(self, gchooks):
        return False

    def remove(self, gchooks):
        return False



from __future__ import print_function
import py
import types
import ast

def d(node):
    import astpp
    import codegen
    print(astpp.dump(node))
    print()
    print(codegen.to_source(node))


class Closure(object):

    def __init__(self, fn, **extravars):
        self.fn = fn
        self.extravars = extravars
        #
        if fn.__closure__:
            for freevar, cell in zip(fn.__code__.co_freevars, fn.__closure__):
                if freevar not in self.extravars:
                    self.extravars[freevar] = cell.cell_contents
        #
        makesrc = self._create_src()
        self.tree = ast.parse(makesrc)
        ast.increment_lineno(self.tree, fn.__code__.co_firstlineno-2)

    def _create_src(self):
        freevars = self.extravars.keys()
        innersrc = py.code.Source(self.fn)
        lines = [
            'def make(%s):' % ', '.join(freevars),
            str(innersrc.indent()),
            '    return %s' % self.fn.__name__
            ]
        return '\n'.join(lines)

    def make(self):
        tree = ast.fix_missing_locations(self.tree)
        co = compile(tree, self.fn.__code__.co_filename, 'exec')
        myglobals = self.fn.__globals__
        mylocals = {}
        exec(co, myglobals, mylocals)
        make = mylocals['make']
        return make(**self.extravars)


def fake_unroll(**kwargs):
    def identity(fn):
        return fn
    return identity


class unroll(object):

    def __init__(self, **extravars):
        self.extravars = tupleify(extravars)
        # we need to specify a fake unroll, to ignore the existing @unroll
        # decorator in the fn source code
        self.extravars['unroll'] = fake_unroll

    def __call__(self, fn):
        closure = Closure(fn, **self.extravars)
        unroller = Unroller(self.extravars)
        closure.tree = unroller.visit(closure.tree)
        return closure.make()


class Unroller(ast.NodeTransformer):
    def __init__(self, extravars):
        self.extravars = extravars

    def visit_For(self, node):
        if isinstance(node.iter, ast.Name) and node.iter.id in self.extravars:
            return self.unroll(node)
        return node

    def unroll(self, fornode):
        assert fornode.orelse == []
        items = self.extravars[fornode.iter.id]
        body = []
        for i in range(len(items)):
            item = ast.Subscript(value=ast.Name(id=fornode.iter.id, ctx=ast.Load()),
                                 slice=ast.Index(value=ast.Num(n=i)),
                                 ctx=ast.Load())
            assign = ast.Assign(targets=[fornode.target],
                                value=item)
            body.append(assign)
            body.extend(fornode.body)
        return body


def tupleify(d):
    """
    Convert to tuple all the iterables in d
    """
    for key, value in d.items():
        try:
            iter(value)
        except TypeError:
            pass
        else:
            d[key] = tuple(value)
    return d



import pytest
from pypytools import IS_PYPY
from pypytools.gc.multihook import MultiHook, GcHooks
from pypytools.gc.testing.test_fakegc import fakegc

@pytest.fixture
def mh(monkeypatch):
    # install itself as MultiHook singleton, to avoid leaving around
    # global state
    obj = MultiHook()
    monkeypatch.setattr(MultiHook, '_instance', obj)
    return obj

class GcStatistics(GcHooks):
    def __init__(self):
        self.reset()

    def reset(self):
        self.minors = []
        self.steps = []
        self.collects = []

    def on_gc_minor(self, stats):
        self.minors.append(stats)

    def on_gc_collect_step(self, stats):
        self.steps.append(stats)

    def on_gc_collect(self, stats):
        self.collects.append(stats)



class TestMultiHook:

    def test_add_remove_hook(self, fakegc):
        class A:
            on_gc_minor = 'A minor'
            on_gc_collect_step = 'A step'
            on_gc_collect = 'A collect'

        class B:
            on_gc_minor = 'B minor'

        mh = MultiHook()
        a = A()
        b = B()
        mh.add(a)
        assert mh.minor_callbacks == ['A minor']
        assert mh.collect_step_callbacks == ['A step']
        assert mh.collect_callbacks == ['A collect']

        mh.add(b)
        assert mh.minor_callbacks == ['A minor', 'B minor']
        assert mh.collect_step_callbacks == ['A step']
        assert mh.collect_callbacks == ['A collect']

        mh.remove(a)
        assert mh.minor_callbacks == ['B minor']
        assert mh.collect_step_callbacks == []
        assert mh.collect_callbacks == []

    def test_install_hook(self, fakegc):
        a = GcStatistics()
        mh = MultiHook()
        mh.add(a)

        fakegc.fire_minor('minor')
        fakegc.fire_step('step')
        fakegc.fire_collect('collect')
        assert a.minors == ['minor']
        assert a.steps == ['step']
        assert a.collects == ['collect']

    def test_dont_install_hook_if_no_cb(self, fakegc):
        class A(object):
            def __init__(self):
                self.minors = []

            def on_gc_minor(self, stats):
                self.minors.append(stats)

        a = A()
        mh = MultiHook()
        mh.add(a)
        assert fakegc.hooks.on_gc_minor is not None
        assert fakegc.hooks.on_gc_collect_step is None
        assert fakegc.hooks.on_gc_collect is None
        fakegc.fire_minor('minor')
        assert a.minors == ['minor']

    def test_check_no_other_hooks(self, fakegc):
        # check that nobody else is messing with gc.hooks directly
        fakegc.hooks.on_gc_minor = lambda stats: None
        with pytest.raises(ValueError):
            MultiHook()

    def test_check_no_other_hooks_when_adding(self, fakegc):
        # check that nobody else is messing with gc.hooks directly
        mh = MultiHook()
        fakegc.hooks.on_gc_minor = lambda stats: None
        with pytest.raises(ValueError):
            mh.add(object())


class TestGcHooks(object):

    def test_mh(self, fakegc, mh):
        assert MultiHook.get() is mh

    def test_enable_disable(self, fakegc, mh):
        class A(GcHooks):
            def __init__(self):
                self.minors = []

            def on_gc_minor(self, stats):
                self.minors.append(stats)

        a1 = A()
        a2 = A()
        a1.enable()
        fakegc.fire_minor(1)
        a2.enable()
        fakegc.fire_minor(2)
        a1.disable()
        fakegc.fire_minor(3)

        assert a1.minors == [1, 2]
        assert a2.minors == [2, 3]

    def test_real_hooks(self, mh):
        # note that:
        #   1. we are NOT using fakegc, so MultiHook uses the builtin real gc mod
        #   2. we run this test also on CPython. The point is that it should
        #      still run without failing
        import gc
        a1 = GcStatistics()
        a2 = GcStatistics()
        a1.enable()
        a2.enable()
        gc.collect()
        if IS_PYPY:
            assert a1.collects[0].count >= 1
            assert a2.collects[0].count >= 1



from pypytools.pypylog import model

class TestEvent(object):

    def test_as_point(self):
        ev = model.Event('123abc', 'foo', 5, 15)
        assert ev.as_point() == (5, 10)


class TestSeries(object):

    def test_from_points(self):
        s = model.Series.from_points([(0, 1), (1, 2), (2, 4), (3, 8)])
        assert len(s) == 4
        assert list(s.X) == [0, 1, 2, 3]
        assert list(s.Y) == [1, 2, 4, 8]

    def test_iter(self):
        plist = [(0, 1), (1, 2), (2, 4), (3, 8)]
        s = model.Series.from_points(plist)
        plist2 = list(s)
        assert plist == plist2 

    def test_get_set_item(self):
        s = model.Series.from_points([(0, 0), (1, 1)])
        assert s[0] == (0, 0)
        assert s[1] == (1, 1)
        s[1] = (2, 2)
        assert s[1] == (2, 2)


def test_make_step_chart():
    events = [
        model.Event('123abc', 'gc',  100, 120),
        model.Event('456abc', 'jit', 130, 135)
    ]
    s = model.make_step_chart(events)
    points = list(s)
    assert points == [
        (100, 0),
        (100, 20),
        
        (100, 20),
        (120, 20),

        (120, 20),
        (120, 0),

        (130, 0),
        (130, 5),

        (130, 5),
        (135, 5),

        (135, 5),
        (135, 0),
    ]



import pytest
from pypytools import IS_PYPY
from pypytools.gc.custom import CustomGc, DefaultGc
from pypytools.gc.testing.test_fakegc import fakegc, FakeMinorStats
from pypytools.gc.testing.test_multihook import GcStatistics, mh

class TestCustomGc:

    def test_enable_disable(self, fakegc):
        class MyGc(CustomGc):

            def __init__(self):
                self.minors = []
                super(MyGc, self).__init__()

            def on_gc_minor(self, stats):
                self.minors.append(stats)

        mygc = MyGc()
        assert not mygc.isenabled()
        assert fakegc.isenabled()
        fakegc.fire_minor('AAA')
        assert mygc.minors == []

        mygc.enable()
        assert mygc.isenabled()
        assert not fakegc.isenabled()
        fakegc.fire_minor('BBB')
        assert mygc.minors == ['BBB']

        mygc.disable()
        assert not mygc.isenabled()
        assert fakegc.isenabled()
        fakegc.fire_minor('CCC')
        assert mygc.minors == ['BBB'] # CCC is not there


class TestDefaultGc:

    def test_threshold(self):
        class MyGc(DefaultGc):
            MAJOR_COLLECT = 2.0
            MIN_THRESHOLD = 100
            MAX_GROWTH = 1.5

        mygc = MyGc()
        assert mygc.threshold == 100

        mygc.update_threshold(mem=200)
        assert mygc.threshold == 150 # capped by MAX_GROWTH

        mygc.update_threshold(mem=80)
        assert mygc.threshold == 160 # mem * MAJOR_COLLECT

    def test_run_steps(self, fakegc):
        class MyGc(DefaultGc):
            MAJOR_COLLECT = 2.0
            MIN_THRESHOLD = 100
            MAX_GROWTH = 100

        def S(mem):
            return FakeMinorStats(total_memory_used=mem)

        stats = GcStatistics()
        mygc = MyGc()
        stats.enable()
        mygc.enable()

        assert mygc.threshold == 100
        fakegc._steps_to_major = 3
        fakegc.fire_minor(S(mem=50))
        assert not mygc.major_in_progress
        assert len(stats.steps) == 0

        fakegc.fire_minor(S(mem=101))
        assert mygc.major_in_progress
        assert len(stats.steps) == 1

        fakegc.fire_minor(S(mem=102))
        assert mygc.major_in_progress
        assert len(stats.steps) == 2

        fakegc.fire_minor(S(mem=103))
        assert not mygc.major_in_progress
        assert len(stats.steps) == 2 # the last collect_step doesn't call the hook
        assert mygc.threshold == 206

        fakegc.fire_minor(S(mem=104))
        assert not mygc.major_in_progress
        assert len(stats.steps) == 2 # no more steps

    def test_nogc(self, fakegc):
        class MyGc(DefaultGc):
            MAJOR_COLLECT = 2.0
            MIN_THRESHOLD = 100
            MAX_GROWTH = 100

        def S(mem):
            return FakeMinorStats(total_memory_used=mem)

        stats = GcStatistics()
        mygc = MyGc()
        stats.enable()
        mygc.enable()

        fakegc.fire_minor(S(mem=101))
        assert len(stats.steps) == 1
        with mygc.nogc():
            fakegc.fire_minor(S(mem=102))
            assert len(stats.steps) == 1 # don't do a collect_step here
        fakegc.fire_minor(S(mem=102))
        assert len(stats.steps) == 2 # collect_step done as usual

    def test_nogc_exception(self, fakegc):
        class MyGc(DefaultGc):
            pass

        mygc = MyGc()
        mygc.enable()
        with pytest.raises(ValueError):
            with mygc.nogc():
                assert mygc.nogc_count == 1
                raise ValueError()
        assert mygc.nogc_count == 0


    @pytest.mark.skip('flaky test')
    @pytest.mark.skipif(not IS_PYPY, reason='PyPy only test')
    def test_real_gc(self):
        import gc # real pypy GC
        from collections import deque
        
        def allocate_some():
            # allocate and discard a total of 100 MB
            d = deque(maxlen=5)
            for i in range(100):
                s = chr(i%256) * (1024*1024) # 1MB
                d.append(s)

        stats = GcStatistics()
        stats.enable()

        # check that allocate_some triggers the GC
        allocate_some()
        assert len(stats.collects) >= 1

        # now, disable the PyPy GC and check that it's not triggered. This
        # works only on the PyPy "gc-disable" branch for now
        mygc = DefaultGc() # custom GC
        gc.disable()
        gc.collect() # start from "fresh"
        stats.reset()

        # the PyPy GC is disabled, no collections
        allocate_some()
        assert len(stats.collects) == 0

        # enable the custom GC and check that it collects
        mygc.enable()
        allocate_some()
        assert len(stats.collects) >= 1

    @pytest.mark.skipif(IS_PYPY, reason='CPython only test')
    def test_dont_disable_on_CPython(self, mh):
        import gc
        # if we are on CPython, we want DefaultGc to do nothing. In
        # particular, it should NOT disable the builtin GC (because else
        # nobody is going to run finalizers)
        mygc = DefaultGc()
        mygc.enable()
        assert gc.isenabled()



"""
Enable the usage of @cython decorators in pure-python code, WITHOUT the
need of having Cython installed.

Cython is a heavy dependency and it does not make sense to install it on PyPy
*only* for 'import cython' to work.

It seems that the only way to convince cython to evaluate the ``@cython.*``
decorators is to have a textual 'import cython' in the source code. So, the
only way to fake a cython module is to patch sys.modules: we use a context
manager to ensure that sys.modules is patched for as little time as needed.

Example usage::

    from pypytools import fakecython
    with fakecython:
        import cython

    @cython.cfunc
    @cython.locals(a=long, b=long)
    @cython.returns(long)
    def foo(a, b):
        return a+b

"""

import sys

def identity(fn):
    return fn


class FakeCython(object):

    __old_cython = None

    def __enter__(self):
        import sys
        self.__old_cython = sys.modules.pop('cython', None)
        sys.modules['cython'] = self

    def __exit__(self, etype, evalue, tb):
        import sys
        if self.__old_cython is None:
            del sys.modules['cython']
        else:
            sys.modules['cython'] = self.__old_cython
            self.__old_cython = None

    compiled = False

    ccall = staticmethod(identity)
    cfunc = staticmethod(identity)

    @staticmethod
    def returns(t):
        return identity

    # at the moment of writing, @cython.except_ is not a feature of standard
    # Cython, but you need my fork to use it:
    # https://github.com/antocuni/cython/tree/pure-except
    @staticmethod
    def except_(ret):
        return identity

    @staticmethod
    def locals(**kwds):
        return identity

    @staticmethod
    def declare(**kwds):
        pass




"""
The goal of this GC strategy is to spread the GC activity as evenly as
possible.
"""
import time
from collections import deque


KB = 1024.0
MB = KB * KB

class UniformGcStrategy(object):

    # MAJOR_COLLECT is roughly equivalent to PYPY_GC_MAJOR_COLLECT. The normal
    # PyPy GC uses PYPY_GC_MAJOR_COLLECT to compute the threshold for when to
    # start a major collection: however, with UniformGc a major collection
    # will take a lot of time to complete, and in the meantime the user
    # program an callocate more.  So, we use MAJOR_COLLECT to compute the
    # value of target_allocated_mem
    MAJOR_COLLECT = 1.82  # roughly PYPY_GC_MAJOR_COLLECT
    MIN_TARGET = 10*MB    # roughly PYPY_GC_MIN

    # if we are using too much memory, try to run a GC step every
    # EMERGENCY_DELAY, and hope to finish soon. This might happen if we
    # suddenly allocate much more than expected, or if a GC cycle takes more
    # than gc_estimated_t. If EMERGENCY_DELAY is too high, the risk is to
    # allocate faster than the GC can free, and have an endless loop. If it's
    # too low, we risk to dedicate too much time to the GC and block important
    # activity in the user program
    EMERGENCY_DELAY = 0.01 # 10 ms

    # if the measured GC time is already bigger than estimated, we simply
    # adjust the estimated
    ESTIMATED_OVERFLOW_FACTOR = 1.1

    def __init__(self, initial_mem):
        self.last_mem = initial_mem   # last known value of used memory
        self.last_t = time.time()     # time of the last tick
        self.alloc_rate = None        # estimated allocation rate, bytes/s

        # the memory allocated by the user program SINCE THE START of the
        # current collection.  Note that it is NOT equivalent to
        # "current_mem-starting_mem", because during the sweeping phases the
        # current_mem shrinks.  We aim to finish the GC cycle before
        # allocated_mem > target_allocated_mem
        # Note that both are adjusted during a start_major()
        self.allocated_mem = 0
        self.target_allocated_mem = 0
        self.target_mem = 1

        # we don't know how much it will take to complete a GC cycle; we just
        # guess reasonable numbers here, they will be automatically adjusted
        # as soon as we complete our first collection
        self.gc_cumul_t = 0.01 # just a random guess
        self.gc_last_step_duration = 0.001 # another random guess
        self.gc_last_step_t = self.last_t

        # number of major collections fully done so far. Not used by anything,
        # just a statistics
        self.n_majors = -1 # increased to 0 when we start_another_major()
        self.start_another_major(initial_mem)


    # ======================================================================
    # Public API
    # ======================================================================

    # in all the code, the parameter "mem" indicates the memory currently used
    def tick(self, mem):
        """
        Regularly called by the user program. Return True if it is time to run a
        GC step.
        """
        try:
            cur_t = time.time()
            self.update_alloc_stats(cur_t, mem)
            if cur_t >= self.get_time_for_next_step():
                return True
            return False
        finally:
            self.record_mem(cur_t, mem)

    def record_gc_step(self, mem, duration, stats):
        """
        Call this AFTER you call gc.collect_step
        """
        cur_t = time.time()
        self.record_mem(cur_t, mem)
        self.gc_cumul_t += duration

        # if our estimate was too low, too bad. Just increase it. The
        # invariant is that gc_cumul_t < gc_estimated_t
        while self.gc_cumul_t >= self.gc_estimated_t:
            self.gc_estimated_t *= self.ESTIMATED_OVERFLOW_FACTOR

        self.gc_steps += 1
        self.gc_last_step_t = cur_t
        if stats.major_is_done:
            self.start_another_major(mem)

    # ======================================================================
    # Private API
    # ======================================================================

    def record_mem(self, cur_t, mem):
        self.last_t = cur_t
        self.last_mem = mem

    def start_another_major(self, mem):
        # we estimate the time needed for a GC for a given target_mem
        k_gc = self.gc_cumul_t / self.target_mem
        self.n_majors += 1
        self.gc_cumul_t = 0
        self.gc_steps = 0
        self.allocated_mem = 0
        self.target_mem = max(mem * self.MAJOR_COLLECT, self.MIN_TARGET)
        self.target_allocated_mem = self.target_mem - mem
        self.gc_estimated_t = k_gc * self.target_mem

    def update_alloc_stats(self, cur_t, mem):
        delta_t = cur_t - self.last_t
        delta_mem = mem - self.last_mem
        cur_alloc_rate = delta_mem / delta_t # bytes/s
        cur_alloc_rate = max(1, cur_alloc_rate) # avoid ZeroDivisionError later
        if self.alloc_rate is None:
            self.alloc_rate = cur_alloc_rate
        else:
            # equivalent to an exponential moving average
            self.alloc_rate = (self.alloc_rate + cur_alloc_rate) / 2.0
        self.allocated_mem += delta_mem

    def get_time_for_next_step(self):
        """
        The goal is to spread GC activity as evenly as possible, i.e.  for any
        given timespan, the GC/user ratio should be roughly the same.

        We know how much time we spent in the GC for the last step, so now we
        let the user program to run enough to keep the desired ratio.

        What is the ratio? We don't know, but we estimate it based on:

          - the total (estimated) time needed for the GC to complete

          - the total (estimated) time we have before we reach the target,
            given the current allocation rate

        The formula is derived as follows:
            let p be the gc/user ration for the complete GC cycle:
                p = total_gc_time / total_time
        
            we want to maintain the ratio for a given step:
                p = (gc_time) / (gc_time + wait_time) ===>
                wait_time = gc_time * (1-p)/p
        """
        if self.allocated_mem >= self.target_allocated_mem:
            return self.gc_last_step_t + self.EMERGENCY_DELAY

        # the invariant is that gc_cumul_t < gc_estimated_t; it is ensured in
        # record_gc_step
        gc_time_left = self.gc_estimated_t - self.gc_cumul_t
        assert gc_time_left > 0, 'bug in record_gc_step?'

        mem_left = self.target_allocated_mem - self.allocated_mem
        time_left = (mem_left / self.alloc_rate) + gc_time_left
        p = gc_time_left / time_left
        wait_t = self.gc_last_step_duration * (1-p)/p
        return self.gc_last_step_t + wait_t



import sys
import gc
from contextlib import contextmanager
from pypytools import IS_PYPY
from pypytools.gc.multihook import GcHooks

KB = 1024.0
MB = KB * KB

class CustomGc(GcHooks):    
    """
    GC with custom logic which is triggered by hooks. Override on_gc_minor to
    implement your custom logic
    """

    def __init__(self):
        self._isenabled = False

    def isenabled(self):
        return self._isenabled

    def enable(self):
        if self._isenabled:
            return
        self._isenabled = True
        if super(CustomGc, self).enable():
            gc.disable()

    def disable(self):
        if not self._isenabled:
            return
        if super(CustomGc, self).disable():
            gc.enable()
        self._isenabled = False



class DefaultGc(CustomGc):
    """
    Custom GC which mimimcs the default logic of incminimark
    """

    MAJOR_COLLECT = 1.82      # same as PYPY_GC_MAJOR_COLLECT
    MIN_THRESHOLD = 4*8 * MB  # same as PYPY_GC_MIN
    MAX_GROWTH = 1.4          # same as PYPY_GC_GROWTH

    def __init__(self):
        super(DefaultGc, self).__init__()
        self.major_in_progress = False
        self.threshold = 0
        self.update_threshold(0)
        self.nogc_count = 0

    def update_threshold(self, mem):
        self.threshold = max(self.MIN_THRESHOLD,
                             min(mem * self.MAJOR_COLLECT,
                                 self.threshold * self.MAX_GROWTH))

    @contextmanager
    def nogc(self):
        try:
            self.nogc_count += 1
            yield self
        finally:
            self.nogc_count -= 1

    def on_gc_minor(self, stats):
        if self.nogc_count > 0:
            return

        # run the hook inside a nogc() section to avoid calling it
        # recursively. Else, it happens the following:
        #   1. a minor collection occurs, and we enter on_gc_minor
        #   2. we reached the threshold and start a major
        #   3. we call gc.collect_step
        #   4. internally, pypy does ANOTHER minor collection
        #   5. we enter AGAIN this hook
        #   6. we are major_in_progress, so we call another gc.collect_step
        #   7. and so on, until we finish the whole collection in one run :-(
        with self.nogc():
            if self.major_in_progress:
                step_stats = gc.collect_step()
                if step_stats.major_is_done:
                    self.major_in_progress = False
                    self.update_threshold(stats.total_memory_used)

            elif stats.total_memory_used > self.threshold:
                self.major_in_progress = True
                gc.collect_step()



import pypytools
from pypytools.compat import micronumpy as unp

eq = unp.array_equal

class TestMicronumpy(object):

    def test_IS_MICRONUMPY(self):
        assert unp.IS_MICRONUMPY == pypytools.IS_PYPY

    def test_all(self):
        a = unp.array([1, 2, 3])
        b = unp.array([1, 2, 0])
        assert unp.all(a)
        assert not unp.all(b)

    def test_any(self):
        a = unp.array([0, 0, 0])
        b = unp.array([0, 0, 1])
        assert not unp.any(a)
        assert unp.any(b)

    def test_argmax(self):
        a = unp.array([100, 101, 102])
        assert unp.argmax(a) == 2

    def test_argmin(self):
        a = unp.array([100, 101, 102])
        assert unp.argmin(a) == 0

    def test_argsort(self):
        a = unp.array([101, 102, 100])
        assert eq(unp.argsort(a), [2, 0, 1])

    def test_choose(self):
        choices = unp.array([[0, 1, 2, 3],
                             [10, 11, 12, 13],
                             [20, 21, 22, 23],
                             [30, 31, 32, 33]])
        out = unp.choose(unp.array([2, 3, 1, 0]), choices)
        assert eq(out, [20, 31, 12,  3])

    def test_clip(self):
        a = unp.array([1, 2, 3, 4, 5, 6])
        assert eq(unp.clip(a, 2, 5), [2, 2, 3, 4, 5, 5])

    def test_copy(self):
        a = unp.array([1, 2, 3])
        b = unp.copy(a)
        assert a is not b
        assert eq(a, b)

    def test_cumprod(self):
        a = unp.array([1, 2, 3, 4])
        assert eq(unp.cumprod(a), [1, 2, 6, 24])

    def test_cumsum(self):
        a = unp.array([1, 2, 3, 4])
        assert eq(unp.cumsum(a), [1, 3, 6, 10])

    def test_diagonal(self):
        a = unp.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 9]])
        assert eq(unp.diagonal(a), [1, 5, 9])

    def test_max(self):
        a = unp.array([1, 2, 3])
        assert unp.max(a) == 3

    def test_min(self):
        a = unp.array([3, 1, 2])
        assert unp.min(a) == 1

    def test_nonzero(self):
        a = unp.array([100, 0, 102, 103])
        (idx,) = unp.nonzero(a)
        assert eq(idx, [0, 2, 3])

    def test_prod(self):
        a = unp.array([1, 2, 3])
        assert unp.prod(a) == 6

    def test_ptp(self):
        a = unp.array([1, 2, 3])
        assert unp.ptp(a) == 2

    def test_put(self):
        a = unp.array([0, 1, 2, 3, 4])
        unp.put(a, [0, 2], [-44, -55])
        assert eq(a, [-44, 1, -55, 3, 4])

    def test_ravel(self):
        a = unp.array([[1, 2, 3],
                       [4, 5, 6]])
        assert eq(unp.ravel(a), [1, 2, 3, 4, 5, 6])

    def test_repeat(self):
        a = unp.array([1, 2, 3])
        assert eq(unp.repeat(a, 2), [1, 1, 2, 2, 3, 3])

    def test_reshape(self):
        a = unp.array([1, 2, 3, 4])
        assert eq(unp.reshape(a, (2, 2)), [[1, 2], [3, 4]])

    def test_round(self):
        a = unp.array([1.1234, 2.4567])
        assert eq(unp.round(a, 2), [1.12, 2.46])

    def test_squeeze(self):
        a = unp.array([[[0], [1], [2]]])
        assert eq(unp.squeeze(a), [0, 1, 2])

    def test_sum(self):
        a = unp.array([1, 2, 3])
        assert unp.sum(a) == 6

    def test_swapaxes(self):
        a = unp.array([[1, 2, 3]])
        b = unp.swapaxes(a, 0, 1)
        assert eq(b, [[1], [2], [3]])

    def test_transpose(self):
        a = unp.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 9]])
        b = unp.transpose(a)
        assert eq(b, [[1, 4, 7],
                      [2, 5, 8],
                      [3, 6, 9]])

    def test_diff(self):
        u8_arr = unp.array([1, 0], dtype=unp.uint8)
        assert eq(unp.diff(u8_arr), [255])
        assert eq(u8_arr[1,...] - u8_arr[0,...], unp.array(255))

        i16_arr = u8_arr.astype(unp.int16)
        assert eq(unp.diff(i16_arr), [-1])

        x = unp.array([1, 2, 4, 7, 0])
        assert eq(unp.diff(x), [1, 2, 3, -7])
        assert eq(unp.diff(x, n=2), [1, 1, -10])

        x = unp.array([[1, 3, 6, 10], [0, 5, 6, 8]])
        assert eq(unp.diff(x), [[2, 3, 4],
                                [5, 1, 2]])
        assert eq(unp.diff(x, axis=0), [[-1,  2,  0, -2]])



import sys
from pypytools import IS_PYPY

def as_signed(x, bits):
    """
    Interpret the given bits-wide int as if it were signed (2-complement
    repr).

    For examplen, given x=0xFF and bits=8:

    bit pattern = 1111 1111
      - unsigned = 255
      - signed   =  -1

    Since in Python ints are always 64 bits (or 32, depending on sys.maxint),
    we need to extend the sign bit up to the most significant bit.

    In CPython, this is done by using simple math; on PyPy, this is done by
    using raw left/right shift operations, which are faster and better handled
    by the JIT.
    """
    if x >= 1<<(bits-1):
        x -= 1<<bits
    return x

if IS_PYPY:
    import __pypy__

    if sys.maxsize == (1<<63)-1:
        BITSIZE = 64
    else:
        BITSIZE = 32

    def as_signed(x, bits):
        shift = BITSIZE - bits
        x = __pypy__.intop.int_lshift(x, shift) # shift x to the most significant bit
        x = __pypy__.intop.int_rshift(x, shift) # shift it back with sign extension
        return x



from pypytools.util import clonefunc

def test_clonefunc():
    def foo(a, b):
        return a+b

    foo2 = clonefunc(foo)
    assert foo is not foo2
    assert foo.__code__ is not foo2.__code__
    assert foo.__code__.co_code is foo2.__code__.co_code
    assert foo2.__name__ == 'foo'
    assert foo2(40, 2) == 42



import re
import attr
from pypytools.pypylog import model

# stolen from rpython/tool/logparse.py
_color = "(?:\x1b.*?m)?"
RE_START = re.compile(_color + r"\[([0-9a-fA-F]+)\] \{([\w-]+)" + _color + "$")
RE_STOP  = re.compile(_color + r"\[([0-9a-fA-F]+)\] ([\w-]+)\}" + _color + "$")

class ParseError(Exception):
    pass


@attr.s
class Section(object):
    tsid = attr.ib()
    name = attr.ib()
    start = attr.ib()
    stop = attr.ib(default=None)
    lines = attr.ib(default=attr.Factory(list))

class BaseParser(object):
    COLLECT_LINES = None

    def __init__(self, log, freq):
        self.log = log
        self.freq = freq

    @classmethod
    def from_file(cls, fname, log=None, freq=1):
        def parse_file(f):
            p = cls(log, freq)
            p.feed(f)
            return log
        #
        if log is None:
            log = model.PyPyLog()
        if isinstance(fname, basestring):
            with open(fname) as f:
                return parse_file(f)
        else:
            return parse_file(fname)

    def parse_line(self, line):
        m = RE_START.match(line)
        if m:
            kind = 'start'
        else:
            m = RE_STOP.match(line)
            if m:
                kind = 'stop'
            else:
                # no start or stop
                return None, None, None, None
        #
        timestamp = m.group(1)
        name = m.group(2)
        return kind, timestamp, self.parse_timestamp(timestamp), name

    def parse_timestamp(self, ts):
        return int(ts, 16) / self.freq

    def feed(self, f):
        zero_ts = None
        stack = []
        for line in f:
            if line == '\n':
                continue
            kind, tsid, ts, name = self.parse_line(line)
            if zero_ts is None:
                zero_ts = ts
            if kind == 'start':
                stack.append(Section(tsid, name, start=ts-zero_ts))
            elif kind == 'stop':
                if name != stack[-1].name:
                    msg = "End section does not match start: expected %s, got %s"
                    raise ParseError(msg % (stack[-1].name, name))
                stack[-1].stop = ts - zero_ts
                self.section(stack[-1])
                stack.pop()
            elif kind is None:
                if self.COLLECT_LINES:
                    stack[-1].lines.append(line)
            else:
                assert False

    def section(self, s):
        pass


class FlatParser(BaseParser):

    def __init__(self, log, freq=1):
        BaseParser.__init__(self, log, freq)

    def section(self, s):
        ev = model.Event(s.tsid, s.name, s.start, s.stop)
        self.log.add_event(ev)


class GcParser(FlatParser):
    COLLECT_LINES = True
    RE_MINOR = re.compile('minor collect, total memory used: ([0-9]*)')
    RE_STEP = re.compile('starting gc state: (.*)')

    def section(self, s):
        name = 'on_%s' % (s.name.replace('-', '_'))
        meth = getattr(self, name, None)
        if meth:
            meth(s)
        else:
            FlatParser.section(self, s)

    def _scan_for_regex(self, regex, lines):
        for line in lines:
            m = regex.match(line)
            if m:
                return m.group(1).strip()
        return None

    def on_gc_minor(self, s):
        memory = self._scan_for_regex(self.RE_MINOR, s.lines)
        if memory:
            memory = int(memory)
        ev = model.GcMinor(s.tsid, s.name, s.start, s.stop, memory=memory)
        self.log.add_event(ev)

    def on_gc_collect_step(self, s):
        phase = self._scan_for_regex(self.RE_STEP, s.lines)
        ev = model.GcCollectStep(s.tsid, s.name, s.start, s.stop, phase=phase)
        self.log.add_event(ev)


flat = FlatParser.from_file
gc = GcParser.from_file

def parse_frequency(s):
    """
    Parse an human-readable string and return the frequency expressed in
    hertz.  It supports different units such as Hz, hz, MHz, GHz, etc.
    """
    s = s.lower().strip()
    if s.endswith('hz'):
        s = s[:-2]
    if not s:
        raise ValueError
    unit = s[-1]
    factor = 1
    if unit == 'k':
        factor = 10**3
    elif unit == 'm':
        factor = 10**6
    elif unit == 'g':
        factor = 10**9
    if factor != 1:
        s = s[:-1]
    return float(s) * factor



from __future__ import print_function
import sys
import dis
import bisect
import linecache
try:
    import pypyjit
except ImportError:
    pass

from pypytools.color import Color
from pypytools.util import PY3

if PY3:
    from io import StringIO
else:
    from cStringIO import StringIO

def disass(obj):
    """
    Very hackish wrapper around dis.dis. Return a dict {addr: human_repr}
    """
    buf = StringIO()
    stdout = sys.stdout
    sys.stdout = buf
    try:
        dis.dis(obj)
    finally:
        sys.stdout = stdout
    #
    res = {}
    lines = buf.getvalue().splitlines()
    for line in lines:
        if len(line) < 10:
            continue
        line = line[10:].strip()
        addr, _ = line.split(' ', 1)
        addr = int(addr)
        res[addr] = line
    return res


class CodePrinter(object):

    def __init__(self):
        self._last = None
        self._indent = 0

    def source(self, filename, lineno):
        line = linecache.getline(filename, lineno)[:-1]
        if line != self._last:
            print(line)
            self._last = line
            self._indent = len(line) - len(line.lstrip())

    def _print(self, s, extra=0):
        indent = ' ' * (self._indent+extra)
        print('%s%s' % (indent, s))

    def bytecode(self, s):
        s = Color.set(Color.lightgray, s)
        self._print(s)

    def llop(self, op):
        s = Color.set(Color.yellow, str(op))
        self._print(s, 4)

class JitView(object):

    _is_hook_installed = False

    def __enter__(self):
        if not JitView._is_hook_installed:
            pypyjit.set_compile_hook(self.on_compile)
            JitView._is_hook_installed = True

    def __exit__(self, etype, evalue, tb):
        pass

    ENTER = __enter__.__code__
    EXIT = __exit__.__code__

    def on_compile(self, info):
        if info.jitdriver_name != 'pypyjit':
            return
        enabled = False
        self.printer = CodePrinter()
        for op in info.operations:
            if op.name == 'debug_merge_point' and op.pycode is self.ENTER:
                enabled = True
            elif op.name == 'debug_merge_point' and op.pycode is self.EXIT:
                enabled = False
            elif op.name == 'label':
                print()
                print(Color.set(Color.green_bg, '-' * 80))
                print()
            elif enabled:
                self._print_op(op)

        pypyjit.set_compile_hook(None)
        JitView._is_hook_installed = False

    def _print_op(self, op):
        if op.name == 'debug_merge_point':
            self._print_debug_merge_point(op)
        else:
            self.printer.llop(op)


    def _find_lineno(self, op):
        linestarts = list(dis.findlinestarts(op.pycode))
        i = bisect.bisect(linestarts, (op.bytecode_no,)) - 1
        if i < 0:
            return -1
        else:
            _, lineno  = linestarts[i]
            return lineno

    def _print_debug_merge_point(self, op):
        bytecodes = disass(op.pycode)
        lineno = self._find_lineno(op)
        opcode = bytecodes.get(op.bytecode_no, '')
        #
        self.printer.source(op.pycode.co_filename, lineno)
        self.printer.bytecode(opcode)





import itertools
from collections import defaultdict
import attr
import numpy as np

@attr.s
class Event(object):
    tsid = attr.ib() # unique identifier for an event
    section = attr.ib()
    start = attr.ib()
    end = attr.ib()

    def as_point(self):
        x = self.start
        y = self.end - self.start
        return x, y

@attr.s
class GcMinor(Event):
    memory = attr.ib(default=None)

@attr.s
class GcCollectStep(Event):
    phase = attr.ib(default=None)


class PyPyLog(object):

    def __init__(self):
        self._events = []

    def add_event(self, ev):
        self._events.append(ev)

    def all_events(self):
        return self._events


class GroupedPyPyLog(object):

    def __init__(self):
        self.sections = defaultdict(list)

    def add_event(self, ev):
        self.sections[ev.section].append(ev)

    def print_summary(self):
        fmt = '%-28s %6s %8s'
        print fmt % ('section', 'n', 'delta')
        print '-'*44
        for name, events in sorted(self.sections.iteritems()):
            total = 0
            for ev in events:
                delta = ev.end - ev.start
                assert delta >= 0
                total += delta
            print fmt % (name, len(events), format(delta, '.4f'))

class Series(object):

    def __init__(self, n, dtype='f'):
        self.X = np.empty(n, dtype=dtype)
        self.Y = np.empty(n, dtype=dtype)

    @classmethod
    def from_points(cls, points):
        res = cls(len(points))
        for i, (x, y) in enumerate(points):
            res.X[i] = x
            res.Y[i] = y
        return res

    @classmethod
    def from_events(cls, events):
        return cls.from_points([ev.as_point() for ev in events])

    def __len__(self):
        assert len(self.X) == len(self.Y)
        return len(self.X)

    def __iter__(self):
        for x, y in itertools.izip(self.X, self.Y):
            yield x, y

    def __getitem__(self, i):
        return self.X[i], self.Y[i]

    def __setitem__(self, i, p):
        x, y = p
        self.X[i] = x
        self.Y[i] = y

def make_step_chart(events):
    """
    Construct a Series which appears like a step chart when you draw it using
    connect='pairs'
    """
    # for each event we draw three lines, i.e. 6 points
    n = len(events)
    s = Series(n*6)
    i = 0
    for ev in events:
        _, h = ev.as_point()
        s[i+0] = (ev.start, 0)
        s[i+1] = (ev.start, h)
        #
        s[i+2] = (ev.start, h)
        s[i+3] = (ev.end, h)
        #
        s[i+4] = (ev.end, h)
        s[i+5] = (ev.end, 0)
        #
        i += 6
    #
    return s



import pytest
import textwrap
from cStringIO import StringIO
from pypytools.pypylog import parse
from pypytools.pypylog import model
from pypytools.pypylog.model import Event, GcMinor, GcCollectStep

class TestFlatParser(object):

    def parse(self, text, log=None):
        text = textwrap.dedent(text)
        f = StringIO(text)
        return parse.flat(f, log)

    def test_simple(self):
        log = self.parse("""
        [ff000] {foo
        [ff123] foo}
        [ff456] {bar
        [ff789] bar}
        """)
        assert log.all_events() == [
            Event('ff000', 'foo', 0x000, 0x123),
            Event('ff456', 'bar', 0x456, 0x789)
        ]

    def test_mismatch(self):
        # raise an error for now, but I think there might be real cases in
        # which this happens
        log = """
        [123] {foo
        [789] {bar
        [456] foo}
        [0ab] bar}
        """
        pytest.raises(parse.ParseError, "self.parse(log)")

    def test_nested(self):
        log = self.parse("""
        [ff000] {foo
        [ff200] {bar
        [ff300] bar}
        [ff400] foo}
        [ff500] {baz
        [ff600] baz}
        """)
        assert log.all_events() == [
            Event('ff200', 'bar', 0x200, 0x300),
            Event('ff000', 'foo', 0x000, 0x400),
            Event('ff500', 'baz', 0x500, 0x600)
        ]

    def test_GroupedPyPyLog(self):
        text = """
        [ff000] {foo
        [ff200] foo}
        [ff300] {bar
        [ff400] bar}
        [ff500] {bar
        [ff600] bar}
        [ff700] {foo
        [ff800] foo}
        """
        log = self.parse(text, model.GroupedPyPyLog())
        assert sorted(log.sections.keys()) == ['bar', 'foo']
        assert log.sections['foo'] == [
            Event('ff000', 'foo', 0x000, 0x200),
            Event('ff700', 'foo', 0x700, 0x800)
        ]
        assert log.sections['bar'] == [
            Event('ff300', 'bar', 0x300, 0x400),
            Event('ff500', 'bar', 0x500, 0x600)
        ]


class TestGcParser(object):
    
    def parse(self, text, log=None):
        text = textwrap.dedent(text)
        f = StringIO(text)
        return parse.gc(f, log)

    def test_gc_minor(self):
        text = """
        [ff000] {gc-minor
        [ff001] {gc-minor-walkroots
        [ff002] gc-minor-walkroots}
        minor collect, total memory used: 1000
        number of pinned objects: 0
        [ff100] gc-minor}
        [ff200] {gc-minor
        [ff201] {gc-minor-walkroots
        [ff202] gc-minor-walkroots}
        minor collect, total memory used: 2000
        number of pinned objects: 0
        [ff300] gc-minor}
        """
        log = self.parse(text, model.GroupedPyPyLog())
        assert log.sections['gc-minor'] == [
            GcMinor('ff000', 'gc-minor', 0x000, 0x100, memory=1000),
            GcMinor('ff200', 'gc-minor', 0x200, 0x300, memory=2000),
        ]

    def test_gc_collect_step(self):
        text = """
        [ff000] {gc-collect-step
        starting gc state:  SCANNING
        stopping, now in gc state:  MARKING
        [ff100] gc-collect-step}
        """
        log = self.parse(text)
        assert log.all_events() == [
            GcCollectStep('ff000', 'gc-collect-step', 0x000, 0x100,
                          phase='SCANNING')
            ]

def test_parse_frequency():
    pf = parse.parse_frequency
    assert pf('40') == 40
    assert pf('40hz') == 40
    assert pf('40Hz') == 40
    assert pf('40 Hz') == 40
    assert pf('40 Hz ') == 40
    assert pf('40 KHz') == 40e3
    assert pf('40 MHz') == 40e6
    assert pf('40 GHz') == 40e9
    pytest.raises(ValueError, "pf('')")



from pypytools.cast import as_signed

def test_as_signed():
    assert as_signed(0xFF, 8) == -1
    assert as_signed(0xFF, 9) == 255
    assert as_signed(0xFFFF, 16) == -1
    assert as_signed(0b1000, 4) == -8




import types
from sys import version_info

PY3 = version_info.major == 3

def clonefunc(f):
    """Deep clone the given function to create a new one.

    By default, the PyPy JIT specializes the assembler based on f.__code__:
    clonefunc makes sure that you will get a new function with a **different**
    __code__, so that PyPy will produce independent assembler. This is useful
    e.g. for benchmarks and microbenchmarks, so you can make sure to compare
    apples to apples.

    Use it with caution: if abused, this might easily produce an explosion of
    produced assembler.
    """
    # first of all, we clone the code object
    co = f.__code__
    args = [co.co_argcount, co.co_nlocals, co.co_stacksize, co.co_flags, co.co_code,
            co.co_consts, co.co_names, co.co_varnames, co.co_filename, co.co_name,
            co.co_firstlineno, co.co_lnotab, co.co_freevars, co.co_cellvars]
    if PY3:
        args.insert(1, co.co_kwonlyargcount)
    co2 = types.CodeType(*args)
    #
    # then, we clone the function itself, using the new co2
    f2 = types.FunctionType(co2, f.__globals__, f.__name__, f.__defaults__, f.__closure__)
    return f2



"""
numpy/numpypy compatibility layer
---------------------------------

Nowadays, the recommended way to install numpy on PyPy is to simply run::

  pip install numpy

It installs the "original" numpy package through cpyext, PyPy's C-API
compatibility layer. This ensures maximum compatibility, but the drawback is
that cpyext is very slow.

PyPy comes also with its own reimplementation of (very) few numpy
functionalities, internally known as `micronumpy`, and accessible through the
builtin `_numpypy` module. Since it is written in RPython it is very well
integrated with the JIT, and it's super fast.

For example, consider the followin trivial benchmark::

    import time
    import numpy as np
    #from pypytools.compat import micronumpy as np

    def main():
        N = 1000000
        a = np.zeros(N)
        t1 = time.time()
        tot = 0
        for item in a:
            tot += item
        t2 = time.time()
        print '%d ms' % ((t2-t1)*1000)

On my machine, I get the following results:

  - CPython:          251 ms
  - PyPy + numpy:    2962 ms
  - PyPy + micronumpy: 36 ms

**WARNING**: as written above, micronumpy might not be 100% compatible with
numpy: you should make sure to test your program both on PyPy and CPython, and
check that you get the same results.

The goal of this module is to make it easier to use micronumpy, if you need
the extra performance; in particular:

  - on PyPy, it rearranges _numpypy in a way which makes it a subset of numpy,
    so that in simple cases you can simply use the pattern `from
    pypytools.compat import micronumpy as np`

  - on CPython, it falls back to numpy. This means that you can simply import
    `pypytools.compat.micronumpy` even on CPython, and it should "just work"
"""

import pypytools
if pypytools.IS_PYPY:
    IS_MICRONUMPY = True
    from pypytools.compat.micronumpy.core import *
    from pypytools.compat.micronumpy.numerictypes import *
    from pypytools.compat.micronumpy.function_base import *

else:
    IS_MICRONUMPY = False
    from numpy import *
    from numpy.core import round, max, min



import sys
IS_PYPY = hasattr(sys, 'pypy_translation_info')

from pypytools.unroll import unroll
from pypytools.codegen import Code
from pypytools.jitview import JitView
from pypytools.util import clonefunc
from pypytools._fakecython import FakeCython

fakecython = FakeCython()



import pypytools
assert pypytools.IS_PYPY

def diff(a, n=1, axis=-1):
    """
    Calculate the n-th discrete difference along the given axis.

    The first difference is given by ``out[n] = a[n+1] - a[n]`` along
    the given axis, higher differences are calculated by using `diff`
    recursively.

    Parameters
    ----------
    a : array_like
        Input array
    n : int, optional
        The number of times values are differenced. If zero, the input
        is returned as-is.
    axis : int, optional
        The axis along which the difference is taken, default is the
        last axis.

    Returns
    -------
    diff : ndarray
        The n-th differences. The shape of the output is the same as `a`
        except along `axis` where the dimension is smaller by `n`. The
        type of the output is the same as the type of the difference
        between any two elements of `a`. This is the same as the type of
        `a` in most cases. A notable exception is `datetime64`, which
        results in a `timedelta64` output array.

    See Also
    --------
    gradient, ediff1d, cumsum

    Notes
    -----
    Type is preserved for boolean arrays, so the result will contain
    `False` when consecutive elements are the same and `True` when they
    differ.

    For unsigned integer arrays, the results will also be unsigned. This
    should not be surprising, as the result is consistent with
    calculating the difference directly:

    >>> u8_arr = np.array([1, 0], dtype=np.uint8)
    >>> np.diff(u8_arr)
    array([255], dtype=uint8)
    >>> u8_arr[1,...] - u8_arr[0,...]
    array(255, np.uint8)

    If this is not desirable, then the array should be cast to a larger
    integer type first:

    >>> i16_arr = u8_arr.astype(np.int16)
    >>> np.diff(i16_arr)
    array([-1], dtype=int16)

    Examples
    --------
    >>> x = np.array([1, 2, 4, 7, 0])
    >>> np.diff(x)
    array([ 1,  2,  3, -7])
    >>> np.diff(x, n=2)
    array([  1,   1, -10])

    >>> x = np.array([[1, 3, 6, 10], [0, 5, 6, 8]])
    >>> np.diff(x)
    array([[2, 3, 4],
           [5, 1, 2]])
    >>> np.diff(x, axis=0)
    array([[-1,  2,  0, -2]])

    >>> x = np.arange('1066-10-13', '1066-10-16', dtype=np.datetime64)
    >>> np.diff(x)
    array([1, 1], dtype='timedelta64[D]')

    """
    from pypytools.compat import micronumpy as np
    if n == 0:
        return a
    if n < 0:
        raise ValueError(
            "order must be non-negative but got " + repr(n))

    a = np.asanyarray(a)
    nd = a.ndim
    #axis = normalize_axis_index(axis, nd)

    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd
    slice1[axis] = slice(1, None)
    slice2[axis] = slice(None, -1)
    slice1 = tuple(slice1)
    slice2 = tuple(slice2)

    op = np.not_equal if a.dtype == np.bool_ else np.subtract
    for _ in range(n):
        a = op(a[slice1], a[slice2])

    return a



class Color:
    black = '30'
    darkred = '31'
    darkgreen = '32'    
    brown = '33'
    darkblue = '34'
    purple = '35'
    teal = '36'
    lightgray = '37'
    darkgray = '30;01'
    red = '31;01'
    green = '32;01'
    yellow = '33;01'
    blue = '34;01'
    fuchsia = '35;01'
    turquoise = '36;01'
    white = '37;01'

    green_bg = '42'

    @classmethod
    def set(cls, color, string):
        try:
            color = getattr(cls, color)
        except AttributeError:
            pass
        return '\x1b[%sm%s\x1b[00m' % (color, string)




#   Copyright 2015-2018 Antonio Cuni anto.cuni@gmail.com

from setuptools import setup

desc = "A collection of useful tools to use PyPy-specific features, with CPython fallbacks"


setup(
    name="pypytools",
    version="0.6.2",
    author="Antonio Cuni",
    author_email="anto.cuni@gmail.com",
    url="http://bitbucket.org/antocuni/pypytools/",
    license="MIT X11 style",
    description=desc,
    packages=["pypytools", "pypytools.compat", "pypytools.gc",
              "pypytools.compat.micronumpy"],
    long_description=desc,
    install_requires=["py"],
)



import pytest
from pypytools.gc import multihook
from pypytools.gc import custom

MODULES_TO_PATCH = [multihook, custom]

@pytest.fixture
def fakegc(monkeypatch):
    mygc = FakeGc()
    for mod in MODULES_TO_PATCH:
        monkeypatch.setattr(mod, 'gc', mygc)
    return mygc


class FakeGc(object):

    class Hooks(object):
        on_gc_minor = None
        on_gc_collect_step = None
        on_gc_collect = None

    def __init__(self):
        self.hooks = self.Hooks()
        self._enabled = True
        self._steps_to_major = 3 # number of collect_step to call before
                                 # finishing a major collection

    def isenabled(self):
        return self._enabled

    def enable(self):
        self._enabled = True

    def disable(self):
        self._enabled = False

    def collect_step(self):
        self._steps_to_major -= 1
        if self._steps_to_major == 0:
            # in PyPy, the last gc.collect_step does not invoke any GC hook,
            # because it runs the app-level finalizers
            self._steps_to_major = 3
            return FakeCollectStepStats(major_is_done=True)
        else:
            stats = FakeCollectStepStats(major_is_done=False)
            # in PyPy, gc.collect_step also does a minor collection, so the effect
            self.fire_minor(FakeMinorStats(total_memory_used=42))
            self.fire_step(stats)
            return stats

    # these are not part of the gc API, but are used to simulate events
    def fire_minor(self, stats):
        if self.hooks.on_gc_minor:
            self.hooks.on_gc_minor(stats)

    def fire_step(self, stats):
        if self.hooks.on_gc_collect_step:
            self.hooks.on_gc_collect_step(stats)

    def fire_collect(self, stats):
        if self.hooks.on_gc_collect:
            self.hooks.on_gc_collect(stats)


class FakeMinorStats(object):

    def __init__(self, total_memory_used):
        self.total_memory_used = total_memory_used


class FakeCollectStepStats(object):

    def __init__(self, major_is_done):
        self.major_is_done = major_is_done


class TestFakeGc:

    def test_fire_hooks(self, fakegc):
        minors = []
        steps = []
        collects = []

        fakegc.fire_minor('minor 1')
        fakegc.fire_step('step 1')
        fakegc.fire_collect('collect 1')

        fakegc.hooks.on_gc_minor = minors.append
        fakegc.hooks.on_gc_collect_step = steps.append
        fakegc.hooks.on_gc_collect = collects.append

        fakegc.fire_minor('minor 2')
        fakegc.fire_step('step 2')
        fakegc.fire_collect('collect 2')

        assert minors == ['minor 2']
        assert steps == ['step 2']
        assert collects == ['collect 2']

    def test_enable_disable(self, fakegc):
        assert fakegc.isenabled()
        fakegc.disable()
        assert not fakegc.isenabled()
        fakegc.enable()
        assert fakegc.isenabled()

    def test_collect_step(self, fakegc):
        steps = []
        fakegc.hooks.on_gc_collect_step = steps.append
        fakegc._steps_to_major = 3
        n = 0
        while True:
            n += 1
            stats = fakegc.collect_step()
            if stats.major_is_done:
                break
            assert n < 100, 'endless loop?'
        assert n == 3
        assert len(steps) == 2 # the last collect_step doesn't call the hook



from pypytools.codegen import Code

def test_build():
    code = Code()
    with code.block('if True:'):
        code.w('x = 42')
    #
    src = code.build()
    assert src == ("if True:\n"
                   "    x = 42")

def test_compile():
    code = Code()
    code.w('x = 42')
    code.compile()
    assert code['x'] == 42

def test_set_get_globals():
    code = Code()
    code['x'] = 42
    assert code['x'] == 42

def test_formatting():
    code = Code()
    code.w('{hello}')
    assert code.build() == '{hello}'
    #
    code = Code()
    code.w('{hello}', hello='world')
    assert code.build() == 'world'

def test_ww():
    code = Code()
    code.ww("""
        if {name}:
            return {name}
    """, name="x")
    src = code.build()
    assert src == ("if x:\n"
                   "    return x")


class TestArgsAndParams(object):
    
    def test_args_simple(self):
        code = Code()
        varnames = ['x', 'y']
        assert code.args(varnames) == 'x, y'
        assert code.args(varnames, '*args') == 'x, y, *args'
        assert code.args(varnames, '*args', '**kwargs') == 'x, y, *args, **kwargs'

    def test_params_simple(self):
        code = Code()
        varnames = ['x', 'y']
        assert code.params(varnames) == 'x, y'
        assert code.params(varnames, '*args') == 'x, y, *args'
        assert code.params(varnames, '*args', '**kwargs') == 'x, y, *args, **kwargs'

    def test_args_default(self):
        code = Code()
        varnames = ['x', ('y', 3)]
        assert code.args(varnames) == 'x, y=3'
        assert code.args(varnames, '*args') == 'x, y=3, *args'
        assert code.args(varnames, '*args', '**kwargs') == 'x, y=3, *args, **kwargs'

    def test_pyx_types(self):
        code = Code(pyx=True)
        varnames = ['x', 'y']
        assert code.args(varnames) == 'x, y'
        assert code.args(varnames, '*args') == 'x, y, *args'
        assert code.params(varnames) == 'object x, object y'
        assert code.params(varnames, '*args') == 'object x, object y, *args'

    def test_pyx_default(self):
        code = Code(pyx=True)
        varnames = ['x', ('y', 3)]
        assert code.params(varnames) == 'object x, object y=3'
        assert code.params(varnames, '*args') == 'object x, object y=3, *args'

    def test_call(self):
        for pyx in (False, True):
            code = Code(pyx=pyx)
            assert code.call('foo', ['x', 'y']) == 'foo(x, y)'


def test_def_():
    code = Code()
    with code.def_('foo', ['x', 'y']):
        code.w('return x+y')
    #
    code.compile()
    foo = code['foo']
    assert foo(1, 2) == 3

def test_def__default_args():
    code = Code()
    with code.def_('foo', ['x', ('y', 3)]):
        code.w('return x+y')
    #
    code.compile()
    foo = code['foo']
    assert foo(1) == 4
    assert foo(1, 10) == 11


def test_cpdef():
    code = Code()
    with code.cpdef_('foo', ['x', 'y']):
        code.w('return x+y')
    src = code.build()
    assert src == ("def foo(x, y):\n"
                   "    return x+y")
    #
    code = Code(pyx=True)
    with code.cpdef_('foo', ['x', 'y']):
        code.w('return x+y')
    src = code.build()
    assert src == ("cpdef foo(object x, object y):\n"
                   "    return x+y")


def test_cdef_var():
    code = Code()
    with code.def_('foo', ['x', 'y']):
        code.cdef_var('long', 'foo')
        code.cdef_var('int', 'bar', 42)
        code.w('return')
    src = code.build()
    assert src == ("def foo(x, y):\n"
                   "    return")
    #
    code = Code(pyx=True)
    with code.def_('foo', ['x', 'y']):
        code.cdef_var('long', 'foo')
        code.cdef_var('int', 'bar', 42)
        code.w('return')
    src = code.build()
    assert src == ("def foo(object x, object y):\n"
                   "    cdef long foo\n"
                   "    cdef int bar = 42\n"
                   "    return")



def test_global_kwargs():
    code = Code()
    code.global_scope.x = 42
    code.w('return {x}')
    assert code.build() == 'return 42'

def test_global_kwargs_override():
    code = Code()
    code.global_scope.x = 42
    code.w('return {x}', x=52)
    assert code.build() == 'return 52'

def test_block_autopass():
    code = Code()
    with code.block('if True:'):
        pass
    src = code.build()
    assert src == ("if True:\n"
                   "    pass")
    #
    code = Code()
    with code.block('if True:', autopass=False):
        pass
    src = code.build()
    assert src == "if True:"

def test_new_global():
    code = Code()
    name = code.new_global('x', 42)
    assert name == 'x'
    assert code['x'] == 42
    #
    name = code.new_global('x', 42)
    assert name == 'x'
    assert code['x'] == 42
    #
    name = code.new_global('x', 43)
    assert name == 'x__1'
    assert code['x'] == 42
    assert code['x__1'] == 43

class TestScope:
    
    def test_new_scope(self):
        code = Code()
        code.global_scope.x = 42
        code.global_scope.y = 52
        ns = code.new_scope(y=53, z=63)
        assert ns.__dict__ == dict(x=42, y=53, z=63)

    def test_scope_formatting(self):
        code = Code()
        ns = code.new_scope(y=52)
        ns.x = 42
        ns.w('x == {x}, y == {y}')
        assert code.build() == 'x == 42, y == 52'

    def test_block_namespace(self):
        code = Code()
        with code.block('if {name}:', name='x') as ns:
            ns.w('return {name}')
        #
        src = code.build()
        assert src == ("if x:\n"
                       "    return x")




"""
Usage: view FILE [options]

Options:
  --dot             Plot the events as dots [default: True]
  --step            Plot the events as steps
  --tsc-freq=FREQ   Convert the TSC counter to seconds with the specified
                    frequency [default: auto]
"""

import sys
from collections import defaultdict
import docopt
import cpuinfo
from pyqtgraph.Qt import QtGui, QtCore
import numpy as np
import pyqtgraph as pg
from pypytools.pypylog import parse
from pypytools.pypylog import model

def make_palette(n):
    """
    Manually run this function if you want to tweak the color
    palette. Normally, the palette is hard-coded, to avoid a dependency on
    matplotlib and seaborn
    """
    from matplotlib.colors import to_hex
    import seaborn as sns
    #palette = sns.hls_palette(n)
    palette = sns.color_palette("Set1", n)
    #palette = sns.color_palette("Dark2", n)
    hex_palette = [str(to_hex(c)) for c in palette]
    print 'PALETTE = %r # %d colors' % (hex_palette, n)
    return hex_palette
    ## import matplotlib
    ## sns.palplot(PALETTE)
    ## matplotlib.pyplot.show()

PALETTE = ['#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00', '#ffff33', '#a65628'] # 7 colors
#PALETTE = make_palette(7)

COLORS = {
    'gc-minor': PALETTE[0],
    'gc-collect-step': PALETTE[1],
    'gc-minor memory': PALETTE[2],

    'jit-optimize': PALETTE[3],
    'jit-backend': PALETTE[4],
    'jit-tracing': PALETTE[5],
    'jit-mem-collect': PALETTE[6],

    # uninteresting sections
    'gc-collect-done': None,
    'gc-hardware': None,
    'gc-minor-walkroots': None,
    'gc-set-nursery-size': None,
    'jit-abort': None,
    'jit-abort-log': None,
    'jit-abort-longest-function': None,
    'jit-backend-addr': None,
    'jit-backend-counts': None,
    'jit-backend-dump': None,
    'jit-disableinlining': None,
    'jit-log-compiling-bridge': None,
    'jit-log-compiling-loop': None,
    'jit-log-exported-state': None,
    'jit-log-noopt': None,
    'jit-log-opt-bridge': None,
    'jit-log-opt-loop': None,
    'jit-log-rewritten-bridge': None,
    'jit-log-rewritten-loop': None,
    'jit-log-short-preamble': None,
    'jit-mem-looptoken-alloc': None,
    'jit-summary': None,
    'jit-trace-done': None,
}

COLORS.update({
    'SCANNING': '#00FF00',
    'MARKING': '#FFFFFF',
    'SWEEPING': '#FFFF00',
    'FINALIZING': '#FF00FF',
    })


class LogViewer(QtCore.QObject):

    def __init__(self, fname, chart_type, freq):
        QtCore.QObject.__init__(self)
        self.global_config()
        self.log = parse.gc(fname, model.GroupedPyPyLog(), freq)
        self.chart_type = chart_type
        self.log.print_summary()
        self.app = pg.mkQApp()
        # main window
        self.win = pg.GraphicsWindow(title=fname)
        self.win.installEventFilter(self) # capture key presses
        self.scene = self.win.scene()
        #
        # plot items, inside the window: we have a different plot item for
        # each Y axis (e.g., one for plotting time and another for potting
        # memory).
        #
        # we create plot_time for last: it's the one which is controlled by
        # dragging the mouse
        self.mem_plot = self.win.addPlot(0, 0)
        self.time_plot = self.win.addPlot(0, 0)
        #
        self.time_plot.setLabel('left', 'Time', 's')
        self.time_plot.setLabel('bottom', 'Time', 's')
        self.time_plot.getAxis('bottom').enableAutoSIPrefix(False)
        #
        self.mem_plot.setXLink(self.time_plot)
        self.mem_plot.showAxis('left', False)
        self.mem_plot.showAxis('bottom', False)
        self.mem_plot.showAxis('right')
        self.mem_plot.setLabel('right', 'Memory', 'B')
        #
        self.time_legend = self.time_plot.addLegend()
        self.mem_legend = self.mem_plot.addLegend(offset=(-30, 30))
        self.make_charts()
        self.add_legend_handlers()
        self.set_axes()

    @staticmethod
    def global_config():
        pg.setConfigOptions(antialias=True)
        pg.setConfigOptions(useOpenGL=True)
        pg.setConfigOption('background', 0.1)
        ## pg.setConfigOption('foreground', 'k')

    def __del__(self):
        self.remove_legend_handlers()

    def show(self):
        self.app.exec_()

    def eventFilter(self, source, event):
        # press ESC to quit
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() in (QtCore.Qt.Key_Escape, ord('Q')):
                self.app.quit()
        return False

    def set_axes(self):
        x_axis = self.time_plot.axes['bottom']['item']
        y_axis = self.time_plot.axes['left']['item']
        x_axis.setGrid(50)
        y_axis.setGrid(50)

    def make_charts(self):
        sections = sorted(self.log.sections)
        for i, name in enumerate(sections):
            if name not in COLORS:
                print 'WARNING: unexpected section %s, ignoring' % name
                COLORS[name] = None
            color = COLORS[name]
            if color is None:
                continue
            events = self.log.sections[name]
            self.make_one_chart(self.chart_type, name, color, events)
        #
        self.make_gc_minor_mem()

    def make_gc_minor_mem(self):
        name = 'gc-minor memory'
        color = COLORS[name]
        events = self.log.sections['gc-minor']
        s = model.Series(len(events))
        for i, ev in enumerate(events):
            s[i] = ev.start, ev.memory
        #print 'Max memory:', s.Y.max()
        self.mem_plot.plot(name=name, x=s.X, y=s.Y, pen=pg.mkPen(color))

    def make_gc_collect_step(self, name, color):
        # draw a plot of gc-collect-step, making sure that each major
        # collection is disconnected from the other ones. We consider a major
        # collection done when we see a transition from FINALIZING to
        # SCANNING.
        pen = pg.mkPen(color)
        steps = self.log.sections['gc-collect-step']
        NaN = float('NaN')
        points = []
        last_ev = None
        phases = defaultdict(list)
        for ev in steps:
            if (last_ev and last_ev.phase == 'FINALIZING' and
                ev.phase == 'SCANNING'):
                points.append((last_ev.start, NaN))
                points.append((ev.start, NaN))
            points.append(ev.as_point())
            last_ev = ev
            phases[ev.phase].append(ev)
        s = model.Series.from_points(points)
        self.time_plot.plot(name=name, x=s.X, y=s.Y, pen=pen, connect='finite')
        #
        # draw points of different colors for each distinct phase
        size = 3
        for phase in ('SCANNING', 'MARKING', 'SWEEPING', 'FINALIZING'):
            color = COLORS[phase]
            self.make_scatter_plot(phase, color, size, phases[phase])

    def make_one_chart(self, t, name, color, events):
        if name == 'gc-collect-step':
            self.make_gc_collect_step(name, color)
        elif t == 'step':
            step_chart = model.make_step_chart(events)
            pen = pg.mkPen(color, width=3)
            self.time_plot.plot(name=name,
                                x=step_chart.X,
                                y=step_chart.Y,
                                connect='pairs',
                                pen=pen)
        elif t == 'dot':
            size = 2
            self.make_scatter_plot(name, color, size, events)
        else:
            raise ValueError('Unknown char type: %s' % t)

    def make_scatter_plot(self, name, color, size, events):
        pen = pg.mkPen(color)
        brush = pg.mkBrush(color)
        size = 2
        s = model.Series.from_events(events)
        item = self.time_plot.scatterPlot(name=name, x=s.X, y=s.Y, size=size,
                                          pen=pen, brush=brush)
        for ev, p in zip(events, item.scatter.points()):
            p.event = ev
        item.sigPointsClicked.connect(self.on_points_clicked)

    def on_points_clicked(self, item, points):
        for p in points:
            print getattr(p, 'event', None)

    def add_legend_handlers(self):
        # toggle visibility of plot by clicking on the legend
        items = (self.time_legend.items +
                 self.mem_legend.items)
        for sample, label in items:
            def clicked(ev, sample=sample, label=label):
                name = label.text
                curve = self.get_curve(name)
                if curve is None:
                    print 'Cannot find curve', name
                    return
                self.set_curve_visibility(curve, sample, label,
                                          not curve.isVisible())
            #
            sample.mouseClickEvent = clicked
            label.mouseClickEvent = clicked

    def remove_legend_handlers(self):
        # delete the mouseClickEvent attributes which were added by
        # add_legend_handlers: if we don't, we get a segfault during shutdown
        # (not sure why)
        items = (self.time_legend.items +
                 self.mem_legend.items)
        for sample, label in items:
            del sample.mouseClickEvent
            del label.mouseClickEvent

    def get_curve(self, name):
        curves = (self.time_plot.curves +
                  self.mem_plot.curves)
        for curve in curves:
            if curve.name() == name:
                return curve
        return None

    def set_curve_visibility(self, curve, sample, label, visible):
        if visible:
            sample.setOpacity(1)
            label.setOpacity(1)
            curve.show()
        else:
            sample.setOpacity(0.5)
            label.setOpacity(0.5)
            curve.hide()

def get_frequency(freq):
    if freq == 'auto':
        freq = cpuinfo.get_cpu_info()['hz_advertised_raw'][0] # Hz
        return float(freq)
    else:
        return parse.parse_frequency(freq)

def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    chart_type = 'dot'
    if args['--step']:
        chart_type = 'step'
    freq = get_frequency(args['--tsc-freq'])
    viewer = LogViewer(args['FILE'], chart_type, freq)
    viewer.show()

if __name__ == '__main__':
    main()



import sys
from pypytools._fakecython import FakeCython
from pypytools.util import PY3

if PY3:
    long = int

class TestFakeCython:

    def test_fakecython(self):
        fake = FakeCython()
        @fake.cfunc
        @fake.locals(a=long, b=long)
        @fake.returns(long)
        def foo(a, b):
            return a+b
        #
        assert foo(20, 22) == 42

    def test_context_manager(self):
        fakecython = FakeCython()
        sys.modules['cython'] = 'foobar'
        import cython
        assert cython == 'foobar'
        #
        with fakecython:
            import cython
            assert isinstance(cython, FakeCython)
        #
        import cython
        assert cython == 'foobar'
